xquery version "3.1";

(:~
 : This module is responsible for retrieving the variants of a given text on a
 : given page. For this, it leverages the tokenization of the text that takes
 : place during commons:get-page-fragment.
 : 
 : The output of the main function, vars:get-variants, is a sequence of maps of
 : the following format:
 :
 : map {
 :  "body" 
 :  map {
 :      "x-content-type" : "Variant",
 :      "value" : [ map {
 :          "entry" : "${text-of-respective-witness}",
 :          "witness" : "${ms-id}"
 :      }],
 :      "format" : "text/plain",
 :      "type" : "TextualBody"
 :    },
 :    "target" :array { map {
 :      "format" : "text/xml",
 :      "language" : "syc",
 :      "id" : "http://ahikar.sub.uni-goettingen.de/ns/annotations/${teixml-uri}/${token-id}"
 :    } },
 :    "type" : "Annotation",
 :    "id" : "http://ahikar.sub.uni-goettingen.de/ns/annotations/${teixml-uri}/annotation-variants-${token-id}"
 : }
 : 
 :)

module namespace vars="http://ahiqar.uni-goettingen.de/ns/annotations/variants";

declare namespace fn="http://www.w3.org/2005/xpath-functions";
declare namespace tei="http://www.tei-c.org/ns/1.0";

import module namespace commons="http://ahiqar.uni-goettingen.de/ns/commons" at "../commons.xqm";

declare variable $vars:ns := "http://ahiqar.uni-goettingen.de/ns/annotations";
(:~
    Main function to create variant annotations for a given page. The page has to be delivered including token IDs.
    These IDs will be looked up in the collation results returning a sequence of rows (since we are dealing with
    kind of tabular data coming from CollateX).
:)
declare function vars:get-variants($pages as element(tei:TEI)+,
    $teixml-uri as xs:string,
    $pb-n as xs:string) as item()* {
    for $page in $pages
    let $ms-id := vars:get-ms-id-from-idno($teixml-uri)
    let $ms-id-in-collation := "t_" || $ms-id
    
    let $collations := collection("/data/collation-results")
    let $tokenIds-on-page := $page//tei:w/@xml:id
    let $collations-for-page := $collations/fn:map/fn:array[@key="table"]/fn:array[.//fn:string[@key eq "id"] = $tokenIds-on-page]
    let $witnesses := $collations-for-page/root()/fn:map/fn:array[@key="witnesses"]/fn:string/text()
    let $log := util:log-system-out("creating variant annotations for document " || $ms-id || " (tei: " || $teixml-uri || ", page " || $pb-n || ")" )
    return
        (: for every token on page we create an annotation of its PHRASE!!! :)
        for $row in $collations-for-page[1]
        where $row//fn:string[@key="id"]/starts-with(., $ms-id-in-collation) = true()
        (: re-assigns the var to go one with a more suiteable format :)
        let $row := vars:parse-row($row, $witnesses)
        
        (: create the target first, as it is the same for all variants representing a token or phrase :)
        let $target :=
            array {
                map {
                    "selector": map {
                        "type": "CssSelector",
                        (: creates a comma-separated list of IDREFs :)
                        "value": ($this?($ms-id-in-collation)?ids?* ! ("#" || .)) => string-join(",")
                    }
                },
                map {
                    "id": $vars:ns || "/" || $teixml-uri || "/"|| $this?($ms-id-in-collation)?ids?*[1] || "-to-" || $this?($ms-id-in-collation)?ids?*[last()],
                    "format": "text/xml",
                    "language": $page//tei:text[@xml:lang]/@xml:lang => string()
                    }, map{
                    "source": "http://example.org/page1", (: todo: source URL :)
                    "selector": map{
                      "type": "TextQuoteSelector",
                      "exact": ($this?($ms-id-in-collation)?strings?*) => string-join(" ")
                    }
                }
            }
        (: iterate over the cells in the row, but omit the cell representing current document :)
        for $witness in map:keys($this)[. != $ms-id-in-collation]
        let $cell := $row?($witness)
        let $string := $cell?strings?* => string-join(" ")
        let $entry := if ($string eq "") then "omisit" else $string
        return
            map{
                "id": $vars:ns || "/" || $teixml-uri || "/annotation-variants-" || $row?($ms-id-in-collation)?ids?*[1] || "-vs-" || $cell?ids?1,
                "type": "Annotation",
                "body": map {
                    "type": "TextualBody",
                    "value": map{
                        "witness": $witness,
                        "entry": $entry
                        },
                    "format": "text/plain",
                    "x-content-type": "Variant"
                    },
                "target": $target
            }
};

(:~
   Create a custom format for a row from CollateX output.
:)
declare function vars:parse-row($row as element(fn:array), $witnesses as xs:string+) as map()+{
    map:merge(
        for $witness at $pos in $witnesses
        let $ids := $row/fn:array[$pos]/fn:map/fn:string[@key eq "id"]/string()
        return
            map{
                $witness: if(not($ids[1])) then () else map{
                    (: "cell": serialize($row/fn:array[$pos], map{"method": "microxml"}), :)
                    "strings": array{ $row/fn:array[$pos]/fn:map/fn:string[@key eq "t"]/string() },
                    "ids": array{ $ids }
                }
            }
    )
};

declare function vars:get-ms-id-from-idno($teixml-uri as xs:string)
as xs:string {
    let $TEI := commons:open-tei-xml($teixml-uri)//tei:TEI
    return
        if (matches($TEI//tei:monogr/tei:editor, "Salhani")) then
            "Salhani"
        else
            commons:make-id-from-idno($TEI)
            => substring-after("t_")
};

